
#include <boost/numeric/bindings/ublas/matrix.hpp>
#include <boost/numeric/bindings/ublas/vector.hpp>
#include <boost/numeric/bindings/lapack/driver/syev.hpp>
#include <boost/numeric/bindings/lapack/driver/gesdd.hpp>
#include <boost/numeric/bindings/ublas/symmetric.hpp>
namespace lapack = boost::numeric::bindings::lapack;
namespace bindings = boost::numeric::bindings;
namespace ublas = boost::numeric::ublas;
#include "jmath/misc.hpp"
#define SYEV 1

#include "image/roi.hpp"



namespace jafar {
namespace image {

	std::ostream& operator<<(std::ostream &os, ConvexRoi::SliceImpl const &c)
		{ os << "(" << c.x << "," << c.w << ")"; return os; }

	ConvexRoi::ConvexRoi(jblas::vec2 x, jblas::sym_mat22 P, double scale)
	{
		hasConvex = true;
		// compute ellipse params
		jblas::vec lambda(2);
#if SYEV
		ublas::matrix<double, ublas::column_major> R(ublas::project(P, ublas::range(0,2), ublas::range(0,2))); 
		ublas::symmetric_adaptor< ublas::matrix<double, ublas::column_major>, ublas::upper > sR(R);
		int ierr = lapack::syev( 'V', sR, lambda);
#else
		ublas::matrix<double, ublas::column_major> R(P); 
		jblas::mat_column_major U(2, 2);
		jblas::mat_column_major VT(2, 2);
		int ierr = lapack::gesdd('A',R,lambda,U,VT);
JFR_DEBUG("U " << U << ", VT " << VT);
		R = U;
#endif
/*jblas::diag_mat ll(2); ll(0,0) = lambda(0); ll(1,1) = lambda(1);
jblas::mat tmp1 = ublas::prod(ll, ublas::trans(R));
jblas::mat tmp2 = ublas::prod(R,tmp1);
JFR_DEBUG("R.L.Rt " << tmp2);*/
		JFR_POSTCOND(ierr==0, "Ellipsoid::Ellipsoid: error in lapack::syev() function, ierr=" << ierr);
		if (lambda(0) < 0) lambda(0) = 0;
		if (lambda(1) < 0) lambda(1) = 0;
		// rename variables
		double a2 = lambda(0)*scale*scale, b2 = lambda(1)*scale*scale;
		double x0 = x(0), y0 = x(1);
		double c = R(0,0), s = R(0,1); // cos and sin t
		double c2 = c*c, s2 = s*s;
// JFR_DEBUG("a2 " << a2 << ", b2 " << b2 << ", R " << R << ", c " << c << ", s " << s << ", c2 " << c2 << ", s2 " << s2);
		// compute ellipse normalized coeffs: http://web.archive.org/web/20041013054628/http:/www.j3d.org/matrix_faq/curvfaq_latest.html#Q21
		// but modified because was not doing the right thing, and with errors...
		double A = b2*c2 + a2*s2;
		double B = b2*s2 + a2*c2;
		double C = -2*x0*(b2*c2+a2*s2) + 2*y0*c*s*(a2-b2);
		double D = -2*y0*(a2*c2+b2*s2) + 2*x0*c*s*(a2-b2);
		double E = 2*c*s*(b2-a2);
		double F = b2*jmath::sqr(c*x0+s*y0) + a2*jmath::sqr(c*y0-s*x0) - a2*b2;

/*		double C = -2*(b2*c*x0 + a2*s*y0);
		double D =  2*(b2*s*x0 - a2*c*y0);
		double E = 2*c*s*(a2-b2);
		double F = b2*x0*x0 + a2*y0*y0 - a2*b2;
*/
// JFR_DEBUG("Equation Ax^2+By^2+Cx+Dy+Exy+F=0 ; A " << A << ", B " << B << ", C " << C << ", D " << D << ", E " << E << ", F " << F);
		// computing ellipse y bounds
		double _a, _b, _c, delta, sdelta; // 2nd order equation coeffs
		double x1, x2, y1, y2;
		int ix1, ix2, iy1, iy2;
		_a = 4*A*B - E*E;
		_b = 4*A*D - 2*C*E;
		_c = 4*A*F - C*C;
		delta = _b*_b - 4*_a*_c;
		if (delta < 0)
		{
			JFR_DEBUG("delta " << delta << ", b " << _b << ", a " << _a << ", c " << _c << ", x " << x << ", P " << P);
			delta = 0;
		}
		sdelta = sqrt(delta);
		y1 = (-_b - sdelta)/(2*_a);
		y2 = (-_b + sdelta)/(2*_a);
		iy1 = std::floor(y1);
		iy2 = std::floor(y2);
		rect.y = iy1;
		rect.height = iy2-iy1+1;
// JFR_DEBUG("rect after y computation " << rect << " sols " << y1 << " and " << y2);
		// compute 2nd degree eq params
		rect.x = 2000000000;
		rect.width = -2000000000;
		initConvex();
		count_ = 0;
		ConvexRoiImpl::iterator it = convex.begin();
		for (int iy = iy1; iy <= iy2; ++iy, ++it)
		{
			// FIXME as we want to round up the area, we should compute at which
			// y are the vertical tangents, and use y+1 before and y after,
			// with separate thres for left and right
			//double y = (iy+0.5 < x(1) ? iy+1 : iy);
			double y;
			if (rect.height == 1) y = iy+0.5; else
			if (rect.height == 2) y = (iy == iy1 ? iy+1 : iy); else
			y = iy1+1 + (iy-iy1)*(rect.height-2)/(double)(rect.height-1);
			_a = A;
			_b = C + E*y;
			_c = B*y*y + D*y + F;
			delta = _b*_b - 4*_a*_c;
			if (delta < 0)
			{
				JFR_DEBUG("delta " << delta << ", b " << _b << ", a " << _a << ", c " << _c << ", x " << x << ", P " << P);
				delta = 0;
			JFR_ASSERT(false, "erreur");
			}
			sdelta = sqrt(delta);
			x1 = (-_b - sdelta)/(2*_a);
			x2 = (-_b + sdelta)/(2*_a);
			ix1 = std::floor(x1);
			ix2 = std::floor(x2);
// JFR_DEBUG("y = " << iy << ", ix1/x1 " << ix1 << "/" << x1 << ", ix2/x2 " << ix2 << "/" << x2);
			it->x = ix1;
			it->w = ix2-ix1+1;
			count_ += it->w;
			if (ix1 < rect.x) rect.x = ix1;
			if (ix2 > rect.width) rect.width = ix2;
		}
		rect.width = rect.width-rect.x+1;
		for(it = convex.begin(); it != convex.end(); ++it)
			it->x -= rect.x;
		checkIntegrity();
	}


	std::ostream& operator<<(std::ostream& os, ConvexRoi const& roi)
	{
		os << roi.rect;
		return os;
	}


	SliceRoi & SliceRoi::operator=(const SliceRoi & roi)
	{
		delete_slices();
		*this = roi;
		SliceRoiImpl::iterator it1;
		SliceRoiImpl::const_iterator it2;
		SliceImpl **slice1, *slice2;
		for(it1 = slices.begin(), it2 = roi.slices.begin(); it2 != roi.slices.end(); ++it1, ++it2)
			for(slice1 = &(*it1), slice2 = *it2 ; slice2 != NULL; slice1 = &((*slice1)->next), slice2 = slice2->next)
				*slice1 = new SliceImpl(slice2->x, slice2->w);
		return *this;
	}

	std::ostream& operator<<(std::ostream& os, SliceRoi const& roi)
	{
		os << roi.rect;
		return os;
	}

	SliceRoi & SliceRoi::operator=(const ConvexRoi & croi)
	{
		rect = croi.rect;
		delete_slices();
		initSlices(true);
		count_ = croi.count_;
		SliceRoi::SliceRoiImpl::iterator it1;
		ConvexRoi::ConvexRoiImpl::const_iterator it2;
		for(it1 = slices.begin(), it2 = croi.convex.begin(); it2 != croi.convex.end(); ++it1, ++it2)
		{
			(*it1)->x = it2->x;
			(*it1)->w = it2->w;
		}
		return *this;
	}


}}

